@@ -29,7 +29,7 @@
 #include <sys/time.h>
 #include <X11/X.h>
 #include <X11/Xos.h>
-#include <X11/Intrinsic.h>
+#include <X11/Xutil.h>
 #include <X11/Xatom.h>
 
 #include "mozplugger.h"
