@@ -32,6 +32,9 @@ static Window victim;
 static Window old_parent;
 static NPWindow windata;
 
+static Atom swallowMutex;
+static int  swallowMutexTaken;
+
 #define WINDOW ((Window) windata.window)
 
 int error_handler(Display *dpy, XErrorEvent *err)
@@ -45,6 +48,170 @@ int error_handler(Display *dpy, XErrorEvent *err)
      return 0;
 }
 
+/*****************************************************************************
+ * Swallow mutex semaphore protection
+ *****************************************************************************/
+static void initSwallowMutex()
+{
+    D("Initialising Swallow Mutex\n"); 
+    if(!display) 
+    {
+	fprintf(stderr, "Display not set so cannot initialise semaphore!\n");
+        return;
+    }
+    swallowMutex = XInternAtom(display, "MOZPLUGGER_SWALLOW_MUTEX", 0);
+    swallowMutexTaken = 0;
+}
+
+unsigned long getHostId()
+{
+    char hostName[128];
+    unsigned long id;
+    int i;
+   
+    memset(hostName, 0, sizeof(hostName));
+    gethostname(hostName, sizeof(hostName)-1);
+
+    D("Host Name = \"%s\"\n", hostName);
+
+    /* OK, Create a 32 bit hash value from the host name....
+       use this as a host ID, the possiblity of a collison of hash keys 
+       effecting the swallow of victims is so infinitimisally small! */
+
+    id = 0;
+    for(i=0; i < sizeof(hostName)/sizeof(unsigned long); i++)
+    {
+        id = ((id << 5) ^ (id >> 27)) ^ ((unsigned long *)hostName)[i];
+    }
+    return id;
+}
+
+static int getSwallowMutexOwner(unsigned long * hostId, unsigned long * pid)
+{   
+    unsigned long nitems;
+    unsigned long bytes;      
+    int fmt;
+    Atom type;
+    unsigned char * property = 0;
+    int success = 0;
+
+    /* Get hold of the Host & PID that current holds the semaphore for 
+       this display! */
+    XGetWindowProperty(display, wattr.root, swallowMutex,
+                       0, 2, 0, XA_INTEGER,
+                       &type, &fmt, &nitems, &bytes,
+                       &property);
+
+    if(property)
+    {
+        /* Just check all is correct! */
+	if((type != XA_INTEGER) || (fmt!=32) || (nitems!=2) || (bytes != 0))
+        {
+	    fprintf(stderr, "XGetWindowProperty returned bad values " 
+                            "%d,%d,%d,%d\n", type, fmt, nitems, bytes);
+        }
+	else
+	{
+            *hostId = ((unsigned long *)property)[0];
+            *pid = ((unsigned long *)property)[1];
+            success = 1;
+        }
+	XFree(property);
+    }
+    return success;
+}
+
+static void setSwallowMutexOwner(unsigned long hostId, unsigned long pid)
+{
+    D("Setting swallow mutex owner, hostId = 0x%08X, pid=%ld\n", hostId, pid); 
+    unsigned long temp[2] = {hostId, pid};
+    XChangeProperty(display, wattr.root, swallowMutex,
+                        XA_INTEGER, 32, PropModeReplace,
+                        (unsigned char*) (&temp), 2);
+}
+ 
+static void takeSwallowMutex()
+{
+    int countDown;
+    unsigned long ourPid = getpid();
+    unsigned long ourHostId = getHostId();
+    unsigned long otherPid;
+    unsigned long otherHostId;
+    unsigned long prevOtherPid;
+ 
+    D("Taking Swallow Mutex\n"); 
+
+    if((display == 0) || (swallowMutex == 0))
+	return;
+
+    /* Try up tp forty times ( 40 * 250ms = 10 seconds) to take 
+       the semaphore... */
+
+    countDown = 40;
+    while(1)
+    {
+        /* While someone owns the semaphore */
+        while(getSwallowMutexOwner(&otherHostId, &otherPid))
+        {
+            if( otherHostId == ourHostId)    
+            {
+                if(otherPid == ourPid)     
+                {
+ 	            /* Great we have either successfully taken the semaphore
+                       OR (unlikely) we previously had the semaphore! Exit
+                       the function...*/
+	            swallowMutexTaken = 1;
+		    return;
+                }
+
+     	    	D("Semaphore currently taken by pid=%ld\n", otherPid);
+ 		/* Check that the process that has the semaphore exists...
+        	   Bit of a hack, I cant find a function to directly check
+                   if process exists. */
+            	if( (getsid(otherPid) < 0 ) && (errno == ESRCH))
+	        {
+	    	    fprintf(stderr,"Strange other Pid(%ld) cannot be found\n", 
+                                   otherPid);
+                    break;    /* OK force early exit of inner while loop */
+                }
+	    }
+
+            /* Check if the owner of semaphore hasn't recently changed if it
+               has restart timer */
+            if(prevOtherPid != otherPid)
+            {
+     	        D("Looks like semaphore's owner has changed pid=%ld\n", 
+                   otherPid);
+	        countDown = 40;
+                prevOtherPid = otherPid;
+            }
+            
+            /* Do one step of the timer... */
+	    countDown--;
+            if(countDown <= 0)
+		break;
+	
+            usleep(250000);        /* 250ms */
+ 	}  
+        /* else no one has semaphore, timeout, or owner is dead -
+           Set us as the owner, but we need to check if we weren't 
+           beaten to it so once more around the loop */
+        setSwallowMutexOwner(ourHostId, ourPid);
+    }
+}
+
+
+static void giveSwallowMutex()
+{
+    D("Giving Swallow Mutex\n"); 
+    if((display == 0) || (swallowMutex == 0) || (swallowMutexTaken ==0))
+	return;
+
+    XDeleteProperty(display, wattr.root, swallowMutex);
+    swallowMutexTaken = 0;
+}
+
+
 /*****************************************************************************
  * Window resizing
  *****************************************************************************/
@@ -343,7 +510,8 @@ static void find_victim(Window window)
 
 	       /* We don't need to listen for new windows anymore */
 	       XSelectInput(display, wattr.root, 0);
-
+	       giveSwallowMutex();
+ 	
 	       if (flags & H_MAXASPECT)
 	       {
 		    XWindowAttributes ca;
@@ -459,12 +627,14 @@ static void check_pipe_fd_events(void)
      {
 	  if (errno == EINTR) return;
 	  D("Winddata read error, exiting\n");
+          giveSwallowMutex();
 	  exit(EX_UNAVAILABLE);
      }
   
      if (n == 0)
      {
 	  D("Winddata EOF, exiting\n");
+          giveSwallowMutex();
 	  exit(EX_UNAVAILABLE);
      }
   
@@ -527,6 +697,7 @@ static void handle_app(int pid)
 	  {
 	       check_all_events();
 	  }
+          giveSwallowMutex();      // Make sure the semaphore has been released
      }
      else
      {
@@ -623,6 +794,19 @@ int main(int argc, char **argv)
 	  }
 	  else
 	  {
+              if(flags & H_SWALLOW)
+              {
+                  /* If we are swallowing a victim window we need to guard
+                     against more than one instance of helper running in
+                     parallel, this is done using a global mutex semaphore. */
+                  initSwallowMutex();
+
+                  /* Belt & braces, make sure we give semaphore on error. */
+                  signal(SIGTERM, giveSwallowMutex);
+
+                  takeSwallowMutex(); 
+               }
+
                /* Has to be done before forking */
 	       save_initial_windows();
 	       XSelectInput(display, WINDOW, SubstructureRedirectMask);
@@ -631,7 +815,10 @@ int main(int argc, char **argv)
 	  }
 
 	  if ((pid = fork()) == -1)
+          {
+               giveSwallowMutex();  
 	       exit(EX_UNAVAILABLE);
+          }
       
 	  if (!pid)
 	  {
